@@ -46,6 +46,7 @@
 #include "dn.h"
 #include "data.h"
 #include "dataprovider.h"
+#include "signkeyjob.h"
 
 #include "t-support.h"
 
@@ -70,9 +71,6 @@ class TestVarious: public QGpgMETest
 {
     Q_OBJECT
 
-Q_SIGNALS:
-    void asyncDone();
-
 private Q_SLOTS:
     void testDN()
     {
@@ -136,9 +134,7 @@ private Q_SLOTS:
 
         auto ctx = Context::createForProtocol(key.protocol());
         QVERIFY (ctx);
-        TestPassphraseProvider provider;
-        ctx->setPassphraseProvider(&provider);
-        ctx->setPinentryMode(Context::PinentryLoopback);
+        hookUpPassphraseProvider(ctx);
 
         QVERIFY(!ctx->addUid(key, uid));
         delete ctx;
@@ -189,9 +185,7 @@ private Q_SLOTS:
 
         auto ctx = Context::createForProtocol(key.protocol());
         QVERIFY (ctx);
-        TestPassphraseProvider provider;
-        ctx->setPassphraseProvider(&provider);
-        ctx->setPinentryMode(Context::PinentryLoopback);
+        hookUpPassphraseProvider(ctx);
 
         // change expiration of the main key
         QVERIFY(!ctx->setExpire(key, 1000));
@@ -229,10 +223,125 @@ private Q_SLOTS:
         std::vector<Subkey> primaryKey;
         primaryKey.push_back(key.subkey(0));
         const auto err = ctx->setExpire(key, 3000, primaryKey);
-        QCOMPARE(err.code(), GPG_ERR_NOT_FOUND);
+        QCOMPARE(err.code(), static_cast<int>(GPG_ERR_NOT_FOUND));
         delete ctx;
     }
 
+    void testSignKeyWithoutExpiration()
+    {
+        Error err;
+
+        if (!loopbackSupported()) {
+            return;
+        }
+
+        auto ctx = Context::create(OpenPGP);
+        QVERIFY(ctx);
+
+        // Get the signing key (alfa@example.net)
+        auto seckey = ctx->key("A0FF4590BB6122EDEF6E3C542D727CC768697734", err, true);
+        QVERIFY(!err);
+        QVERIFY(!seckey.isNull());
+
+        // Get the target key (Bob / Bravo Test)
+        auto target = ctx->key("D695676BDCEDCC2CDD6152BCFE180B1DA9E3B0B2", err, false);
+        QVERIFY(!err);
+        QVERIFY(!target.isNull());
+        QVERIFY(target.numUserIDs() > 0);
+
+        // Create the job
+        auto job = std::unique_ptr<SignKeyJob>{openpgp()->signKeyJob()};
+        QVERIFY(job);
+        hookUpPassphraseProvider(job.get());
+
+        // Setup the job
+        job->setExportable(true);
+        job->setSigningKey(seckey);
+        job->setDupeOk(true);
+
+        connect(job.get(), &SignKeyJob::result,
+                this, [this] (const GpgME::Error &err2, const QString &, const GpgME::Error &) {
+                    Q_EMIT asyncDone();
+                    if (err2) {
+                        if (err2.code() == GPG_ERR_GENERAL) {
+                            QFAIL(qPrintable(QString("The SignKeyJob failed with '%1'.\n"
+                                "Hint: Run with GPGMEPP_INTERACTOR_DEBUG=stderr to debug the edit interaction.").arg(err2.asString())));
+                        } else {
+                            QFAIL(qPrintable(QString("The SignKeyJob failed with '%1'.").arg(err2.asString())));
+                        }
+                    }
+                });
+
+        job->start(target);
+        QSignalSpy spy{this, &TestVarious::asyncDone};
+        QVERIFY(spy.wait(QSIGNALSPY_TIMEOUT));
+
+        // At this point the signature should have been added.
+        target.update();
+        const auto keySignature = target.userID(0).signature(target.userID(0).numSignatures() - 1);
+        QVERIFY(keySignature.neverExpires());
+    }
+
+    void testSignKeyWithExpiration()
+    {
+        Error err;
+
+        if (!loopbackSupported()) {
+            return;
+        }
+
+        auto ctx = Context::create(OpenPGP);
+        QVERIFY(ctx);
+
+        // Get the signing key (alfa@example.net)
+        auto seckey = ctx->key("A0FF4590BB6122EDEF6E3C542D727CC768697734", err, true);
+        QVERIFY(!err);
+        QVERIFY(!seckey.isNull());
+
+        // Get the target key (Bob / Bravo Test)
+        auto target = ctx->key("D695676BDCEDCC2CDD6152BCFE180B1DA9E3B0B2", err, false);
+        QVERIFY(!err);
+        QVERIFY(!target.isNull());
+        QVERIFY(target.numUserIDs() > 0);
+
+        // Create the job
+        auto job = std::unique_ptr<SignKeyJob>{openpgp()->signKeyJob()};
+        QVERIFY(job);
+        hookUpPassphraseProvider(job.get());
+
+        // Setup the job
+        job->setExportable(true);
+        job->setSigningKey(seckey);
+        job->setDupeOk(true);
+        job->setExpirationDate(QDate{2222, 2, 22});
+
+        connect(job.get(), &SignKeyJob::result,
+                this, [this] (const GpgME::Error &err2, const QString &, const GpgME::Error &) {
+                    Q_EMIT asyncDone();
+                    if (err2) {
+                        if (err2.code() == GPG_ERR_GENERAL) {
+                            QFAIL(qPrintable(QString("The SignKeyJob failed with '%1'.\n"
+                                "Hint: Run with GPGMEPP_INTERACTOR_DEBUG=stderr to debug the edit interaction.").arg(err2.asString())));
+                        } else {
+                            QFAIL(qPrintable(QString("The SignKeyJob failed with '%1'.").arg(err2.asString())));
+                        }
+                    }
+                });
+
+        QTest::ignoreMessage(QtWarningMsg, "Expiration of certification has been changed to QDate(\"2106-02-06\")");
+
+        job->start(target);
+        QSignalSpy spy{this, &TestVarious::asyncDone};
+        QVERIFY(spy.wait(QSIGNALSPY_TIMEOUT));
+
+        // At this point the signature should have been added.
+        target.update();
+        const auto keySignature = target.userID(0).signature(target.userID(0).numSignatures() - 1);
+        QVERIFY(!keySignature.neverExpires());
+        const auto expirationDate = QDateTime::fromSecsSinceEpoch(uint_least32_t(keySignature.expirationTime())).date();
+        QCOMPARE(expirationDate, QDate(2106, 2, 6));  // expiration date is capped at 2106-02-06
+    }
+
     void testVersion()
     {
         QVERIFY(EngineInfo::Version("2.1.0") < EngineInfo::Version("2.1.1"));
@@ -285,6 +394,12 @@ private Q_SLOTS:
         const QString gpgHome = qgetenv("GNUPGHOME");
         QVERIFY(copyKeyrings(gpgHome, mDir.path()));
         qputenv("GNUPGHOME", mDir.path().toUtf8());
+        QFile conf(mDir.path() + QStringLiteral("/gpg.conf"));
+        QVERIFY(conf.open(QIODevice::WriteOnly));
+        if (GpgME::engineInfo(GpgME::GpgEngine).engineVersion() >= "2.2.18") {
+            conf.write("allow-weak-key-signatures");
+        }
+        conf.close();
     }
 
 private:
