@@ -15,11 +15,11 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, see <http://www.gnu.org/licenses/>.
+ * along with this program; if not, see <https://www.gnu.org/licenses/>.
+ * SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #include <config.h>
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -29,7 +29,7 @@
 #include <unistd.h>
 #include <assert.h>
 
-#include "i18n.h"
+#include "../common/i18n.h"
 #include "../common/util.h"
 #include "../common/asshelp.h"
 #include "../common/sysutils.h"
@@ -44,6 +44,9 @@
 #define CONTROL_D ('D' - 'A' + 1)
 #define octdigitp(p) (*(p) >= '0' && *(p) <= '7')
 
+#define HISTORYNAME ".gpg-connect_history"
+
+
 /* Constants to identify the commands and options. */
 enum cmd_and_opt_values
   {
@@ -60,18 +63,23 @@ enum cmd_and_opt_values
     oHomedir,
     oAgentProgram,
     oDirmngrProgram,
+    oKeyboxdProgram,
     oHex,
     oDecode,
     oNoExtConnect,
     oDirmngr,
+    oKeyboxd,
     oUIServer,
+    oNoHistory,
     oNoAutostart,
+    oChUid,
 
+    oNoop
   };
 
 
 /* The list of commands and options. */
-static ARGPARSE_OPTS opts[] = {
+static gpgrt_opt_t opts[] = {
   ARGPARSE_group (301, N_("@\nOptions:\n ")),
 
   ARGPARSE_s_n (oVerbose, "verbose", N_("verbose")),
@@ -79,6 +87,7 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_s_n (oHex,   "hex",       N_("print data out hex encoded")),
   ARGPARSE_s_n (oDecode,"decode",    N_("decode received data lines")),
   ARGPARSE_s_n (oDirmngr,"dirmngr",  N_("connect to the dirmngr")),
+  ARGPARSE_s_n (oKeyboxd,"keyboxd",  N_("connect to the keyboxd")),
   ARGPARSE_s_n (oUIServer, "uiserver", "@"),
   ARGPARSE_s_s (oRawSocket, "raw-socket",
                 N_("|NAME|connect to Assuan socket NAME")),
@@ -94,9 +103,13 @@ static ARGPARSE_OPTS opts[] = {
 
   ARGPARSE_s_n (oNoAutostart, "no-autostart", "@"),
   ARGPARSE_s_n (oNoVerbose, "no-verbose", "@"),
+  ARGPARSE_s_n (oNoHistory,"no-history",
+                "do not use the command history file"),
   ARGPARSE_s_s (oHomedir, "homedir", "@" ),
   ARGPARSE_s_s (oAgentProgram, "agent-program", "@"),
   ARGPARSE_s_s (oDirmngrProgram, "dirmngr-program", "@"),
+  ARGPARSE_s_s (oKeyboxdProgram, "keyboxd-program", "@"),
+  ARGPARSE_s_s (oChUid,          "chuid",           "@"),
 
   ARGPARSE_end ()
 };
@@ -109,11 +122,13 @@ struct
   int quiet;		/* Be extra quiet.  */
   int autostart;        /* Start the server if not running.  */
   const char *homedir;  /* Configuration directory name */
-  const char *agent_program;  /* Value of --agent-program.  */
+  const char *agent_program;    /* Value of --agent-program.    */
   const char *dirmngr_program;  /* Value of --dirmngr-program.  */
+  const char *keyboxd_program;  /* Value of --keyboxd-program.  */
   int hex;              /* Print data lines in hex format. */
   int decode;           /* Decode received data lines.  */
   int use_dirmngr;      /* Use the dirmngr and not gpg-agent.  */
+  int use_keyboxd;      /* Use the keyboxd and not gpg-agent.  */
   int use_uiserver;     /* Use the standard UI server.  */
   const char *raw_socket; /* Name of socket to connect in raw mode. */
   const char *tcp_socket; /* Name of server to connect in tcp mode. */
@@ -121,6 +136,7 @@ struct
   unsigned int connect_flags;    /* Flags used for connecting. */
   int enable_varsubst;  /* Set if variable substitution is enabled.  */
   int trim_leading_spaces;
+  int no_history;
 } opt;
 
 
@@ -166,7 +182,7 @@ typedef struct loopline_s *loopline_t;
 static pid_t server_pid = (pid_t)(-1);
 
 /* The current datasink file or NULL.  */
-static FILE *current_datasink;
+static estream_t current_datasink;
 
 /* A list of open file descriptors. */
 static struct
@@ -187,7 +203,7 @@ static assuan_context_t start_agent (void);
 
 
 
-/* Print usage information and and provide strings for help. */
+/* Print usage information and provide strings for help. */
 static const char *
 my_strusage( int level )
 {
@@ -195,9 +211,11 @@ my_strusage( int level )
 
   switch (level)
     {
+    case  9: p = "GPL-3.0-or-later"; break;
     case 11: p = "@GPG@-connect-agent (@GNUPG@)";
       break;
     case 13: p = VERSION; break;
+    case 14: p = GNUPG_DEF_COPYRIGHT_LINE; break;
     case 17: p = PRINTABLE_OS_NAME; break;
     case 19: p = _("Please report bugs to <@EMAIL@>.\n"); break;
 
@@ -413,7 +431,7 @@ get_var (const char *name)
 }
 
 
-/* Perform some simple arithmentic operations.  Caller must release
+/* Perform some simple arithmetic operations.  Caller must release
    the return value.  On error the return value is NULL.  */
 static char *
 arithmetic_op (int operator, const char *operands)
@@ -486,7 +504,7 @@ arithmetic_op (int operator, const char *operands)
 
      unescape ARGS
            Remove C-style escapes from string.  Note that "\0" and
-           "\x00" terminate the string implictly.  Use "\x7d" to
+           "\x00" terminate the string implicitly.  Use "\x7d" to
            represent the closing brace.  The args start right after
            the first space after the function name.
 
@@ -839,7 +857,7 @@ add_definq (char *line, int is_var, int is_prog)
 }
 
 
-/* Show all inquiry defintions. */
+/* Show all inquiry definitions. */
 static void
 show_definq (void)
 {
@@ -874,7 +892,7 @@ clear_definq (void)
 static void
 do_sendfd (assuan_context_t ctx, char *line)
 {
-  FILE *fp;
+  estream_t fp;
   char *name, *mode, *p;
   int rc, fd;
 
@@ -900,14 +918,14 @@ do_sendfd (assuan_context_t ctx, char *line)
     }
 
   /* Open and send. */
-  fp = fopen (name, mode);
+  fp = es_fopen (name, mode);
   if (!fp)
     {
       log_error ("can't open '%s' in \"%s\" mode: %s\n",
                  name, mode, strerror (errno));
       return;
     }
-  fd = fileno (fp);
+  fd = es_fileno (fp);
 
   if (opt.verbose)
     log_error ("file '%s' opened in \"%s\" mode, fd=%d\n",
@@ -916,7 +934,7 @@ do_sendfd (assuan_context_t ctx, char *line)
   rc = assuan_sendfd (ctx, INT2FD (fd) );
   if (rc)
     log_error ("sending descriptor %d failed: %s\n", fd, gpg_strerror (rc));
-  fclose (fp);
+  es_fclose (fp);
 }
 
 
@@ -932,7 +950,7 @@ do_recvfd (assuan_context_t ctx, char *line)
 static void
 do_open (char *line)
 {
-  FILE *fp;
+  estream_t fp;
   char *varname, *name, *mode, *p;
   int fd;
 
@@ -976,14 +994,14 @@ do_open (char *line)
     }
 
   /* Open and send. */
-  fp = fopen (name, mode);
+  fp = es_fopen (name, mode);
   if (!fp)
     {
       log_error ("can't open '%s' in \"%s\" mode: %s\n",
                  name, mode, strerror (errno));
       return;
     }
-  fd = fileno (fp);
+  fd = dup (es_fileno (fp));
   if (fd >= 0 && fd < DIM (open_fd_table))
     {
       open_fd_table[fd].inuse = 1;
@@ -1030,8 +1048,10 @@ do_open (char *line)
   else
     {
       log_error ("can't put fd %d into table\n", fd);
-      close (fd);
+      if (fd != -1)
+        close (fd); /* Table was full.  */
     }
+  es_fclose (fp);
 }
 
 
@@ -1146,7 +1166,7 @@ help_cmd_p (const char *line)
 int
 main (int argc, char **argv)
 {
-  ARGPARSE_ARGS pargs;
+  gpgrt_argparse_t pargs;
   int no_more_options = 0;
   assuan_context_t ctx;
   char *line, *p;
@@ -1155,7 +1175,7 @@ main (int argc, char **argv)
   int rc;
   int cmderr;
   const char *opt_run = NULL;
-  FILE *script_fp = NULL;
+  gpgrt_stream_t script_fp = NULL;
   int use_tty, keep_line;
   struct {
     int collecting;
@@ -1168,11 +1188,16 @@ main (int argc, char **argv)
   } loopstack[20];
   int        loopidx;
   char **cmdline_commands = NULL;
+  char *historyname = NULL;
+
+  static const char *changeuser;
+
 
   early_system_init ();
   gnupg_rl_initialize ();
-  set_strusage (my_strusage);
-  log_set_prefix ("gpg-connect-agent", GPGRT_LOG_WITH_PREFIX);
+  gpgrt_set_strusage (my_strusage);
+  log_set_prefix ("gpg-connect-agent",
+                  GPGRT_LOG_WITH_PREFIX|GPGRT_LOG_NO_REGISTRY);
 
   /* Make sure that our subsystems are ready.  */
   i18n_init();
@@ -1180,6 +1205,7 @@ main (int argc, char **argv)
 
   assuan_set_gpg_err_source (0);
 
+  gnupg_init_signals (0, NULL);
 
   opt.autostart = 1;
   opt.connect_flags = 1;
@@ -1187,8 +1213,8 @@ main (int argc, char **argv)
   /* Parse the command line. */
   pargs.argc  = &argc;
   pargs.argv  = &argv;
-  pargs.flags =  1;  /* Do not remove the args.  */
-  while (!no_more_options && optfile_parse (NULL, NULL, NULL, &pargs, opts))
+  pargs.flags = ARGPARSE_FLAG_KEEP;
+  while (!no_more_options && gpgrt_argparse (NULL, &pargs, opts))
     {
       switch (pargs.r_opt)
         {
@@ -1198,10 +1224,13 @@ main (int argc, char **argv)
         case oHomedir:   gnupg_set_homedir (pargs.r.ret_str); break;
         case oAgentProgram: opt.agent_program = pargs.r.ret_str;  break;
         case oDirmngrProgram: opt.dirmngr_program = pargs.r.ret_str;  break;
+        case oKeyboxdProgram: opt.keyboxd_program = pargs.r.ret_str;  break;
         case oNoAutostart:    opt.autostart = 0; break;
+        case oNoHistory: opt.no_history = 1; break;
         case oHex:       opt.hex = 1; break;
         case oDecode:    opt.decode = 1; break;
         case oDirmngr:   opt.use_dirmngr = 1; break;
+        case oKeyboxd:   opt.use_keyboxd = 1; break;
         case oUIServer:  opt.use_uiserver = 1; break;
         case oRawSocket: opt.raw_socket = pargs.r.ret_str; break;
         case oTcpSocket: opt.tcp_socket = pargs.r.ret_str; break;
@@ -1212,10 +1241,15 @@ main (int argc, char **argv)
           opt.enable_varsubst = 1;
           opt.trim_leading_spaces = 1;
           break;
+        case oChUid:     changeuser = pargs.r.ret_str; break;
 
         default: pargs.err = 2; break;
 	}
     }
+  gpgrt_argparse (NULL, &pargs, NULL);  /* Release internal state.  */
+
+  if (changeuser && gnupg_chuid (changeuser, 0))
+    log_inc_errorcount (); /* Force later termination.  */
 
   if (log_get_errorcount (0))
     exit (2);
@@ -1271,7 +1305,7 @@ main (int argc, char **argv)
                 "--tcp-socket", "--raw-socket");
     }
 
-  if (opt_run && !(script_fp = fopen (opt_run, "r")))
+  if (opt_run && !(script_fp = gpgrt_fopen (opt_run, "r")))
     {
       log_error ("cannot open run file '%s': %s\n",
                  opt_run, strerror (errno));
@@ -1409,6 +1443,15 @@ main (int argc, char **argv)
         {
           keep_line = 0;
           xfree (line);
+          if (!historyname && !opt.no_history)
+            {
+              historyname = make_filename (gnupg_homedir (), HISTORYNAME, NULL);
+              if (tty_read_history (historyname, 500))
+                log_info ("error reading '%s': %s\n",
+                          historyname,
+                          gpg_strerror (gpg_error_from_syserror ()));
+            }
+
           line = tty_get ("> ");
           n = strlen (line);
           if (n==1 && *line == CONTROL_D)
@@ -1425,15 +1468,15 @@ main (int argc, char **argv)
               linesize = 0;
               keep_line = 1;
             }
-          n = read_line (script_fp? script_fp:stdin,
-                         &line, &linesize, &maxlength);
+          n = gpgrt_read_line (script_fp ? script_fp : gpgrt_stdin,
+                               &line, &linesize, &maxlength);
         }
       if (n < 0)
         {
           log_error (_("error reading input: %s\n"), strerror (errno));
           if (script_fp)
             {
-              fclose (script_fp);
+              gpgrt_fclose (script_fp);
               script_fp = NULL;
               log_error ("stopping script execution\n");
               continue;
@@ -1445,7 +1488,7 @@ main (int argc, char **argv)
           /* EOF */
           if (script_fp)
             {
-              fclose (script_fp);
+              gpgrt_fclose (script_fp);
               script_fp = NULL;
               if (opt.verbose)
                 log_info ("end of script\n");
@@ -1563,17 +1606,17 @@ main (int argc, char **argv)
 
               if (current_datasink)
                 {
-                  if (current_datasink != stdout)
-                    fclose (current_datasink);
+                  if (current_datasink != es_stdout)
+                    es_fclose (current_datasink);
                   current_datasink = NULL;
                 }
               tmpline = opt.enable_varsubst? substitute_line (p) : NULL;
               fname = tmpline? tmpline : p;
               if (fname && !strcmp (fname, "-"))
-                current_datasink = stdout;
+                current_datasink = es_stdout;
               else if (fname && *fname)
                 {
-                  current_datasink = fopen (fname, "wb");
+                  current_datasink = es_fopen (fname, "wb");
                   if (!current_datasink)
                     log_error ("can't open '%s': %s\n",
                                fname, strerror (errno));
@@ -1683,17 +1726,17 @@ main (int argc, char **argv)
                   log_error ("syntax error in run command\n");
                   if (script_fp)
                     {
-                      fclose (script_fp);
+                      gpgrt_fclose (script_fp);
                       script_fp = NULL;
                     }
                 }
               else if (script_fp)
                 {
                   log_error ("cannot nest run commands - stop\n");
-                  fclose (script_fp);
+                  gpgrt_fclose (script_fp);
                   script_fp = NULL;
                 }
-              else if (!(script_fp = fopen (p, "r")))
+              else if (!(script_fp = gpgrt_fopen (p, "r")))
                 {
                   log_error ("cannot open run file '%s': %s\n",
                              p, strerror (errno));
@@ -1796,7 +1839,7 @@ main (int argc, char **argv)
                     }
                 }
             }
-          else if (!strcmp (cmd, "bye"))
+          else if (!strcmp (cmd, "bye") || !strcmp (cmd, "quit"))
             {
               break;
             }
@@ -1804,6 +1847,15 @@ main (int argc, char **argv)
             {
               gnupg_sleep (1);
             }
+          else if (!strcmp (cmd, "history"))
+            {
+              if (!strcmp (p, "--clear"))
+                {
+                  tty_read_history (NULL, 0);
+                }
+              else
+                log_error ("Only \"/history --clear\" is supported\n");
+            }
           else if (!strcmp (cmd, "help"))
             {
               puts (
@@ -1830,6 +1882,7 @@ main (int argc, char **argv)
 "/if VAR                Begin conditional block controlled by VAR.\n"
 "/while VAR             Begin loop controlled by VAR.\n"
 "/end                   End loop or condition\n"
+"/history               Manage the history\n"
 "/bye                   Terminate gpg-connect-agent.\n"
 "/help                  Print this help.");
             }
@@ -1864,7 +1917,7 @@ main (int argc, char **argv)
       if ((rc || cmderr) && script_fp)
         {
           log_error ("stopping script execution\n");
-          fclose (script_fp);
+          gpgrt_fclose (script_fp);
           script_fp = NULL;
         }
 
@@ -1877,7 +1930,16 @@ main (int argc, char **argv)
     }
 
   if (opt.verbose)
-    log_info ("closing connection to agent\n");
+    log_info ("closing connection to %s\n",
+              opt.use_dirmngr? "dirmngr" :
+              opt.use_keyboxd? "keyboxd" :
+              "agent");
+
+
+  if (historyname && tty_write_history (historyname))
+    log_info ("error writing '%s': %s\n",
+              historyname, gpg_strerror (gpg_error_from_syserror ()));
+
 
   /* XXX: We would like to release the context here, but libassuan
      nicely says good bye to the server, which results in a SIGPIPE if
@@ -1888,6 +1950,7 @@ main (int argc, char **argv)
     assuan_release (ctx);
   else
     gpgrt_annotate_leaked_object (ctx);
+  xfree (historyname);
   xfree (line);
   return 0;
 }
@@ -1902,6 +1965,7 @@ handle_inquire (assuan_context_t ctx, char *line)
 {
   const char *name;
   definq_t d;
+  /* FIXME: Due to the use of popen we can't easily switch to estream.  */
   FILE *fp = NULL;
   char buffer[1024];
   int rc, n;
@@ -2053,7 +2117,7 @@ read_and_print_response (assuan_context_t ctx, int withhash, int *r_goterr)
                     }
                   else
                     c = *s;
-                  putc (c, current_datasink);
+                  es_putc (c, current_datasink);
                 }
             }
           else if (opt.hex)
@@ -2124,7 +2188,7 @@ read_and_print_response (assuan_context_t ctx, int withhash, int *r_goterr)
         {
           if (need_lf)
             {
-              if (!current_datasink || current_datasink != stdout)
+              if (!current_datasink || current_datasink != es_stdout)
                 putchar ('\n');
               need_lf = 0;
             }
@@ -2133,7 +2197,7 @@ read_and_print_response (assuan_context_t ctx, int withhash, int *r_goterr)
               && line[0] == 'S'
               && (line[1] == '\0' || line[1] == ' '))
             {
-              if (!current_datasink || current_datasink != stdout)
+              if (!current_datasink || current_datasink != es_stdout)
                 {
                   fwrite (line, linelen, 1, stdout);
                   putchar ('\n');
@@ -2143,7 +2207,7 @@ read_and_print_response (assuan_context_t ctx, int withhash, int *r_goterr)
                    && line[0] == 'O' && line[1] == 'K'
                    && (line[2] == '\0' || line[2] == ' '))
             {
-              if (!current_datasink || current_datasink != stdout)
+              if (!current_datasink || current_datasink != es_stdout)
                 {
                   fwrite (line, linelen, 1, stdout);
                   putchar ('\n');
@@ -2161,7 +2225,7 @@ read_and_print_response (assuan_context_t ctx, int withhash, int *r_goterr)
               if (!errval)
                 errval = -1;
               set_int_var ("?", errval);
-              if (!current_datasink || current_datasink != stdout)
+              if (!current_datasink || current_datasink != es_stdout)
                 {
                   fwrite (line, linelen, 1, stdout);
                   putchar ('\n');
@@ -2175,7 +2239,7 @@ read_and_print_response (assuan_context_t ctx, int withhash, int *r_goterr)
                    && line[6] == 'E'
                    && (line[7] == '\0' || line[7] == ' '))
             {
-              if (!current_datasink || current_datasink != stdout)
+              if (!current_datasink || current_datasink != es_stdout)
                 {
                   fwrite (line, linelen, 1, stdout);
                   putchar ('\n');
@@ -2187,7 +2251,7 @@ read_and_print_response (assuan_context_t ctx, int withhash, int *r_goterr)
                    && line[0] == 'E' && line[1] == 'N' && line[2] == 'D'
                    && (line[3] == '\0' || line[3] == ' '))
             {
-              if (!current_datasink || current_datasink != stdout)
+              if (!current_datasink || current_datasink != es_stdout)
                 {
                   fwrite (line, linelen, 1, stdout);
                   putchar ('\n');
@@ -2222,6 +2286,13 @@ start_agent (void)
                              opt.autostart,
                              !opt.quiet, 0,
                              NULL, NULL);
+  else if (opt.use_keyboxd)
+    err = start_new_keyboxd (&ctx,
+                             GPG_ERR_SOURCE_DEFAULT,
+                             opt.keyboxd_program,
+                             opt.autostart,
+                             !opt.quiet, 0,
+                             NULL, NULL);
   else
     err = start_new_gpg_agent (&ctx,
                                GPG_ERR_SOURCE_DEFAULT,
@@ -2237,12 +2308,15 @@ start_agent (void)
     {
       if (!opt.autostart
           && (gpg_err_code (err)
-              == opt.use_dirmngr? GPG_ERR_NO_DIRMNGR : GPG_ERR_NO_AGENT))
+              == (opt.use_dirmngr? GPG_ERR_NO_DIRMNGR :
+                  opt.use_keyboxd? GPG_ERR_NO_KEYBOXD : GPG_ERR_NO_AGENT)))
         {
           /* In the no-autostart case we don't make gpg-connect-agent
              fail on a missing server.  */
           log_info (opt.use_dirmngr?
                     _("no dirmngr running in this session\n"):
+                    opt.use_keyboxd?
+                    _("no keybox daemon running in this session\n"):
                     _("no gpg-agent running in this session\n"));
           exit (0);
         }
